"""
this file is to bring datasets into specific representations
"""
import pandas as pd
from pyadlml.util import get_parallel, get_npartitions
from pyadlml.constants import OTHER, TIME, ACTIVITY, START_TIME, END_TIME

def label_data(df_devs: pd.DataFrame, df_acts: pd.DataFrame, other=False, n_jobs=1, inplace=True):
    """
    Label a dataframe with corresponding activities based on a time-index.

    Parameters
    ----------
    df_devs : pd.DataFrame
        some data representation that possesses a column 'time' including timestamps.
    df_acts : pd.DataFrame
        a datasets activities. TODO
    other : bool, optional, default=False
        if true this leads to datapoints not falling into a logged activity to be
        labeled as "other"
    n_jobs : int, optional, default=1
        the number of jobs that are run in parallel TODO look up sklearn
    inplace : bool, optional, default=True
        determines whether a new column is appended to the existing dataframe.

    Examples
    --------
    >>> raw = DiscreteEncoder()
    >>> raw
    1 time                    0   ...      13
    2 2008-03-20 00:34:38  False  ...    True
    3 2008-03-20 00:34:39  False  ...   False

    now include
    >>> label_data(raw, data.df_activities, other=True, n_jobs=10)
    1 time                    0   ...      13 activity
    2 2008-03-20 00:34:38  False  ...    True other
    3 2008-03-20 00:34:39  False  ...   False act1

    Returns
    -------
    df : pd.DataFrame
    """

    df_devs = df_devs.copy()
    df_devs[ACTIVITY] = -1

    if n_jobs == 1:
        df_devs[ACTIVITY] = df_devs[TIME].apply(
                    _map_timestamp2activity,
                    df_act=df_acts,
                    other=other)
    else:
        N = get_npartitions()
        if n_jobs == -1 or n_jobs > N:
            n_jobs = N

        import dask.dataframe as dd
        #ddf_activities = dd.from_pandas(df_activities, npartitions=get_npartitions())
        # compute with dask in parallel
        df_devs[ACTIVITY] = dd.from_pandas(df_devs[TIME], npartitions=n_jobs).\
                    map_partitions( # apply lambda functions on each partition
                        lambda df: df.apply(
                            _map_timestamp2activity,
                            df_act=df_acts,
                            other=other)).\
                    compute(scheduler='processes')
    return df_devs



def _map_timestamp2activity(timestamp, df_act, other):
    """ Map the given timestamp map to an activity in df_act

    Parameters
    ----------
    timestamp : pd.Timestamp
        E.g timestamp 2008-02-26 00:39:25
    df_act : pd.DataFrame
        An activity dataframe
    other : boolean
        Whether to map gaps to NAT or "other"

    Returns
    -------
        label of the activity
    """

    # select activity intervals that the timestamp falls into
    mask = (df_act[START_TIME] <= timestamp) & (timestamp <= df_act[END_TIME])
    matches = df_act[mask]
    match_amount = len(matches.index)

    # 1. case no activity interval matched
    if match_amount == 0 and other:
        return OTHER
    elif match_amount == 0 and not other:
        return pd.NaT

    # 2. case single row matches
    elif match_amount == 1:
        return matches.activity.values[0]
    
    # 3. case multiple rows
    else:
        print()
        print('*'*70)
        print('ts: ', timestamp)
        print('matches: ', matches)
        print('matches_amount: ', match_amount)
        print('overlap of activities. this should be handled when loading activities')
        print('*'*70)
        raise ValueError




from torch import nn
from pyadlml.model.nn import Tanh, Embedding
from torch.nn import BatchNorm1d, Sequential



class MLP(nn.Module):
    def __init__(self, n_features, window_length, n_embed, n_hidden, n_classes):
        super().__init__()
        self.layers = Sequential( 
            Embedding(n_features, n_embed),         # B, T      -> B, T, emb
            nn.Flatten(),                           # B, T      -> B, T*emb

            nn.Linear(n_embed*window_length, n_hidden, bias=False),
            BatchNorm1d(n_hidden), 
            Tanh(),

            nn.Linear(n_hidden, n_hidden, bias=False),
            BatchNorm1d(n_hidden), 
            Tanh(),

            nn.Linear(n_hidden, n_hidden, bias=False),
            BatchNorm1d(n_hidden), 
            Tanh(),

            #nn.Linear(n_hidden, n_hidden, bias=False), BatchNorm1d(n_hidden), Tanh(),
            nn.Linear(n_hidden, n_classes),         # B, H      -> B, C
        )

    #def parameters(self):
    #    return self.layers
    #    #return [p for layer in self.layers for p in layer.parameters()]

    def __call__(self, X):
        for l in self.layers:
            X = l(X)
        return X




import pandas as pd
from sklearn.base import TransformerMixin, BaseEstimator
from pyadlml.constants import TIME, DEVICE

def _tres_2_pd_res(res):
    if _is_hour(res) or _is_sec(res):
        return res
    if _is_min(res):
        return res[:-1] + 'min'


def _tres_2_discrete(resolution):
    """ create resolution for one day
    """
    res = []
    range = pd.date_range(start='1/1/2000', end='1/2/2000', freq=resolution).time
    range = [str(d) for d in range]
    for t1, t2 in zip(range[:-1], range[1:]):
        res.append(t1 + '-' + t2)
    return res


def _tres_2_vals(resolution):
    """ create resolution for one day
    """
    ser = pd.date_range(start='1/1/2000', end='1/2/2000', freq=resolution).to_series()
    if _is_min(resolution):
        return ser.dt.floor(resolution).dt.minute
    if _is_hour(resolution): return ser.dt.floor(resolution).dt.hour
    if _is_sec(resolution):
        return ser.dt.floor(resolution).dt.sec


def _is_hour(res) -> bool:
    return res[-1:] == 'h'


def _is_min(res) -> bool:
    return res[-1:] == 'm'


def _is_sec(res) -> bool:
    return res[-1:] == 's'


def _valid_tres(t_res):
    val = int(t_res[:-1])
    res = t_res[-1:]
    assert t_res[-1:] in ['h', 'm', 's']
    assert (val > 0 and val <= 12 and res == 'h') \
           or (val > 0 and val <= 60 and res == 'm') \
           or (val > 0 and val <= 60 and res == 's')
    return True


def _time2intervall(ts, t_res='30m'):
    """
    rounds to the next lower min bin or hour bin """
    ts_ceil = ts.ceil(freq=t_res).time()
    ts_floor = ts.floor(freq=t_res).time()
    return str(ts_floor) + '-' + str(ts_ceil)


def extract_time_difference(df, normalize=False, inplace=True):
    return df


class DayOfWeekExtractor(TransformerMixin, BaseEstimator):
    """
    Appends seven columns one-hot encoded for week days to a dataframe based
        on the dataframes timestamps.

    Attributes
    ----------
    one_hot_encoding : bool, optional, default=False
        Determines

    inplace : bool, default=True
        Determines whether to append the one-hot encoded time_bins as columns
        to the existing dataframe or return only the one-hot encoding.

    """

    WD_COL = 'weekday'
    WEEKDAYS = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']

    def __init__(self, one_hot_encoding=False, inplace=False):
        self.one_hot_encoding = one_hot_encoding
        self.inplace = inplace

    def fit(self, X):
        return self

    def fit_transform(self, X, y=None, **fit_params):
        return self.transform(X)

    def transform(self, X):
        """
        Appends seven columns one-hot encoded for week days to a dataframe based
        on the dataframes timestamps.

        Parameters
        ----------
        X : pd.DataFrame
            A device dataframe with a column named 'time' containing timestamps.

        Returns
        ------
        df : pd.DataFrame
        """

        df = X.copy()
        df[self.WD_COL] = df[TIME].dt.day_name()
        if not self.one_hot_encoding:
            if self.inplace:
                return df
            else:
                return df[self.WD_COL]

        one_hot = pd.get_dummies(df[self.WD_COL])
        df = df.join(one_hot, on=df.index)

        # add weekdays that didn't occur in the column
        for wd in self.WEEKDAYS:
            if wd not in df.columns:
                df[wd] = False
        del (df[self.WD_COL])
        if self.inplace:
            return df
        else:
            return df[self.WEEKDAYS]


class TimeBinExtractor(TransformerMixin, BaseEstimator):
    """
    Divides a day into equal-length time bins and assigns time bin as features to
    a row that falls into that bin.

    Attributes
    ----------
    dt: str, default='2h'
        The frequency that the day is divided. Different resolutions are
        accepted, like minutes '[x]m' seconds '[x]s] or hours '[x]h'
    inplace : bool, default=False
        Determines whether to append the one-hot encoded time_bins as columns
        to the existing dataframe or return only the one-hot encoding.

    """
    TIME_BINS = 'time_bins'
    ALLOWED_RESOLUTIONS = ['m', 's', 'h']

    def __init__(self, dt='2h', one_hot_encoding=False, inplace=False):
        self.dt = dt
        self.inplace = inplace
        self.one_hot_encoding = one_hot_encoding

    def fit(self, X, y=None):
        assert self.dt[-1:] in self.ALLOWED_RESOLUTIONS
        return self

    def fit_transform(self, X, y=None, **fit_params):
        """

        Parameters
        ----------
        X: pd.DataFrame
           A device dataframe. The dataframe has to include a column with
           the name 'time' containing the timestamps of the representation.

        Returns
        -------
        df : pd.DataFrame
            One-hot encoding of the devices.
        """
        self.fit(X,y)

        return self.transform(X)

    def transform(self, X, y=None):
        """

        Parameters
        ----------
        X: pd.DataFrame
           A device dataframe. The dataframe has to include a column
           named 'time' containing timestamps of the representation.

        Returns
        -------
        df : pd.DataFrame
            One-hot encoding of the devices.
        """
        assert self.dt[-1:] in self.ALLOWED_RESOLUTIONS

        df = X.copy()
        df[self.TIME_BINS] = df[TIME].apply(_time2intervall, args=[self.dt])

        if not self.one_hot_encoding:
            if self.inplace:
                return df
            else:
                return df[self.TIME_BINS]

        one_hot = pd.get_dummies(df[self.TIME_BINS]).astype(bool)
        df = df.join(one_hot, on=df.index)
        del (df[self.TIME_BINS])

        # add columns that don't exist in the dataset
        cols = _tres_2_discrete(self.dt)
        for v in cols:
            if v not in df.columns:
                df[v] = False

        if self.inplace:
            return df
        else:
            return df[cols]


class EventTimeExtractor(TransformerMixin, BaseEstimator):
    """
    Extracts the time difference to either the next or previous event

    Attributes
    ----------
    direction : str, optional, default='to_predecessor'
        The direction to which the timedifference is calculated.
    normalize : bool, optional, default=False
        Whether to apply min-max scaling to normalize the time differences to the interval [0,1]
    inplace : bool, optional , default=False
        Determines whether to add the column to and return the existing dataframe or
        return a dataframe containing only the time differences.
    unit : one of {None, 'm', 's'}, default=None
        The unit in which the time difference is expressed. When nothing is specified
        transform returns pandas timedeltas otherwise integers
    """
    BACKWARD = 'to_predecessor'
    FORWARD = 'to_successor'
    COL_NAME = 'td'

    def __init__(self, direction='to_predecessor', unit=None, normalize=False, inplace=False):
        self.normalize = normalize
        self.inplace = inplace
        self.direction = direction
        self.unit = unit

    def _compute_td(self, df):
        """
        compute the time difference to successor and predecessor
        """
        df[self.FORWARD] = df[TIME].shift(-1) - df[TIME]
        df[self.BACKWARD] = df[TIME] - df[TIME].shift(1)
        return df

    def fit(self, X: pd.DataFrame, y=None):
        assert TIME in X.columns

        # compute a dataframe with mean time diff for each device
        df = self._compute_td(X.copy())

        # compute average value
        self.td_avg = df[[DEVICE, self.BACKWARD, self.FORWARD]]\
            .iloc[1:-1]\
            .groupby(DEVICE)\
            .mean(numeric_only=False)\
            .reset_index()

        # compute the highest and lowest value
        if self.normalize:
            self.min_ = df[self.direction].max()
            self.max_ = df[self.direction].min()

        return self

    def fit_transform(self, X, y=None, **fit_params):
        assert TIME in X.columns

        self.fit(X)
        return self.transform(X)

    def transform(self, X, y=None):
        """
        Adds one column with time difference between two succeeding rows.

        Parameters
        ----------
        X : pd.DataFrame
            A dataframe containing a column named 'time' including valid pandas timestamps.

        Returns
        -------
        df : pd.DataFrame
        """
        assert TIME in X.columns

        df = self._compute_td(X.copy())

        # impute either last or first value with the average of the value
        if self.direction == self.BACKWARD:
            del(df[self.FORWARD])
            first_device = df.iloc[0][DEVICE]
            df.at[0, self.BACKWARD] = self.td_avg[self.td_avg[DEVICE] == first_device][self.BACKWARD].iloc[0]
            df.rename(columns={self.BACKWARD: self.COL_NAME}, inplace=True)

        elif self.direction == self.FORWARD:
            del(df[self.BACKWARD])
            last_device = df.iloc[-1][DEVICE]
            df.at[0, self.FORWARD] = self.td_avg[self.td_avg[DEVICE] == last_device][self.FORWARD].iloc[0]
            df.rename(columns={self.FORWARD: self.COL_NAME}, inplace=True)
        else:
            raise ValueError

        if self.normalize:
            # apply min max normalization
            df[self.COL_NAME] = (df[self.COL_NAME] - self.min_)/(self.max_ - self.min_)

            # clip values that don't fit
            df[self.COL_NAME] = df[self.COL_NAME].where(df[self.COL_NAME] < 1, 1)
            df[self.COL_NAME] = df[self.COL_NAME].where(df[self.COL_NAME] > 0, 0)

        elif self.unit is not None:
            if self.unit == 's':
                df[self.COL_NAME] = df[self.COL_NAME].dt.seconds
            elif self.unit == 'ms':
                df[self.COL_NAME] = df[self.COL_NAME].dt.microseconds
            elif self.unit == 'h':
                df[self.COL_NAME] = df[self.COL_NAME].dt.hours
            else:
                raise ValueError

        if self.inplace:
            return df
        else:
            return df[self.COL_NAME]


import datetime

import ssm

from hassbrain_algorithm.models._model import Model
import numpy as np
from hassbrain_algorithm.models.hmm.bhmm import BernoulliHMM
from hassbrain_algorithm.models.hmm.bhsmm import BHSMM
from hassbrain_algorithm.models.hmm.util import gen_handcrafted_priors_for_pi, gen_handcrafted_priors_for_transitions

class BernoulliHMM_HandcraftedPriors(BernoulliHMM):

    def __init__(self, controller):
        # c is the constant that is added if an obs is 0
        self._emission_constant = 0.00001
        self._transition_constant = 0.000001
        self._pi_constant = 0.000001
        BernoulliHMM.__init__(self, controller)

    def _model_init(self, dataset):
        state_list = dataset.get_state_list()
        observation_list = dataset.get_obs_list()

        K = len(self._state_list)       # number of discrete states
        D = len(self._observation_list) # dimension of the observation
        self._hmm = ssm.HMM(K, D, observations='bernoulli')

        # note:
        # the log classes govern their own data type
        # therefore set_pi for example transforms matrix to Prob domain itsself
        # no action is needed
        assert self._act_data != None and self._act_data != {}
        init_pi = gen_handcrafted_priors_for_pi(self._act_data, K)
        self.set_new_pi(init_pi)

        # todo this is not evaluated
        #assert self._loc_data != None and self._loc_data != {}
        #em_matrix = self.gen_pc_emissions(K, D)
        #self._hmm.set_emission_matrix(em_matrix)

        assert self._act_data != None and self._act_data != {}
        trans_mat = gen_handcrafted_priors_for_transitions(self._act_data, K)
        self.set_trans_matrix(trans_mat)

    def set_trans_matrix(self, trans_mat):
        #tm = self._hmm.transitions.log_Ps
        log_new_trans = np.log(trans_mat)
        self._hmm.transitions.log_ps = log_new_trans

    def set_new_pi(self, new_pi):
        log_new_pi = np.log(new_pi)
        self._hmm.init_state_distn.log_pi0 = log_new_pi
        #self._hmm.init_state_distn.params(
        #    log_new_pi
        #)


class BernoulliHSMM_HandcraftedPriors(BHSMM):

    def __init__(self, controller):
        # c is the constant that is added if an obs is 0
        self._emission_constant = 0.00001
        self._transition_constant = 0.000001
        self._pi_constant = 0.000001
        BHSMM.__init__(self, controller)

    def _model_init(self, dataset):
        state_list = dataset.get_state_list()
        observation_list = dataset.get_obs_list()

        K = len(self._state_list)       # number of discrete states
        D = len(self._observation_list) # dimension of the observation
        self._hmm = ssm.HMM(K, D, observations='bernoulli')

        # note:
        # the log classes govern their own data type
        # therefore set_pi for example transforms matrix to Prob domain itsself
        # no action is needed
        assert self._act_data != None and self._act_data != {}
        init_pi = gen_handcrafted_priors_for_pi(self._act_data, K)
        self.set_new_pi(init_pi)

        # todo this is not evaluated
        #assert self._loc_data != None and self._loc_data != {}
        #em_matrix = self.gen_pc_emissions(K, D)
        #self._hmm.set_emission_matrix(em_matrix)

        #assert self._act_data != None and self._act_data != {}
        #trans_mat = gen_handcrafted_priors_for_transitions(self._act_data, K)
        #self.set_trans_matrix(trans_mat)

    def set_trans_matrix(self, trans_mat):
        #tm = self._hmm.transitions.log_Ps
        log_new_trans = np.log(trans_mat)
        self._hmm.transitions.log_ps = log_new_trans

    def set_new_pi(self, new_pi):
        log_new_pi = np.log(new_pi)
        self._hmm.init_state_distn.log_pi0 = log_new_pi
        #self._hmm.init_state_distn.params(
        #    log_new_pi
        #)



import datetime

from hbhmm.hmm._hmm_base import HMM
from hbhmm.hmm.hmm_log import HMM_log
from hbhmm.hmm.hmm_log_scaled import HMM_log_scaled
from hbhmm.hmm.hmm_scaled import HMM_Scaled
from hbhmm.hmm.distributions import ProbabilityMassFunction
from hassbrain_algorithm.models._model import Model
from hassbrain_algorithm.models.hmm.hmm import ModelHMM_log_scaled
import numpy as np


class PreConfHMM(ModelHMM_log_scaled):
    def __init__(self, controller):
        # c is the constant that is added if an obs is 0
        self._emission_constant = 0.00001
        self._transition_constant = 0.000001
        self._pi_constant = 0.000001

        ModelHMM_log_scaled.__init__(self, controller)

    def _model_init(self, dataset):
        state_list = dataset.get_state_list()
        observation_list = dataset.get_obs_list()
        K = len(state_list)
        D = len(observation_list)
        self._hmm = HMM_log_scaled(state_list,
                            observation_list,
                            ProbabilityMassFunction,
                            initial_dist=None)

        # note:
        # the log classes govern their own data type
        # therefore set_pi for example transforms matrix to Prob domain itsself
        # no action is needed
        init_pi = self.gen_pc_pi(K)
        self._hmm.set_pi(init_pi)

        em_matrix = self.gen_pc_emissions(K, D)
        self._hmm.set_emission_matrix(em_matrix)

        trans_mat = self.gen_pc_transitions(K)
        self._hmm.set_transition_matrix(trans_mat)


    def gen_pc_pi(self, K):
        """
        calculate the probabilities of beeing in a specific state
        based on the total time spent in the state in the
        synthetic data
        :param K:
            the amount of states :return:
            1D K long numpy array
        """
        if self._act_data == []:
            return HMM_log_scaled.gen_rand_pi(K)

        # total amount of hours spent in this activity
        time_deltas = np.zeros((K), dtype='f')
        total_hours = 0.0
        for act_data_point in self._act_data:
            # todo inconsistent use of typecast
            tidx = self._hmm._idx_state(str(act_data_point['name']))
            td = self._time_diff(act_data_point['end'], act_data_point['start'])
            if act_data_point['end'] == datetime.time.fromisoformat("00:00:00"):
                # if the endtime is 0 o clock the timedifferenz is negativ
                # and has to be corrected
                # example 19:00:00 to 00:00:00 yields -19.0 and corrected is 5.0
                td = td + 24.0
            time_deltas[tidx] += td
            total_hours += td

        new_pi = time_deltas/total_hours # type: np.ndarray
        """
        assign a constant where the probability is 0.0 and correct the other
        probabilities for it
        """
        cnt_non_zero = np.count_nonzero(new_pi)
        cnt_zero = K - cnt_non_zero
        if cnt_zero > 0:
            non_zero_correction = cnt_zero*self._pi_constant/cnt_non_zero
            for i in range(K):
                if new_pi[i] == 0.0:
                    new_pi[i] = self._pi_constant
                else:
                    new_pi[i] -= non_zero_correction
        return new_pi

    def _time_diff(self, a, b):
        """
        computes the differences of two time points
        stackoverflow copy paste
        :param a:
        :param b:
        :return:
            the total amount of time in hours
        """
        dateTimeA = datetime.datetime.combine(datetime.date.today(), a)
        dateTimeB = datetime.datetime.combine(datetime.date.today(), b)

        # Get the difference between datetimes (as timedelta)
        dateTimeDifference = dateTimeA - dateTimeB

        # Divide difference in seconds by number of seconds in hour (3600)
        dateTimeDifferenceInHours = dateTimeDifference.total_seconds() / 3600
        return dateTimeDifferenceInHours


    def gen_pc_transitions(self, K):
        """
        :param K:
            the amount of states
        :return:
            K x K numpy matrix
        """
        sorted_act_data = self.sort_act_data()
        sorted_acts = []
        for act in sorted_act_data:
            sorted_acts.append(act['name'])
        tm = np.zeros((K,K))

        # the total transitions from state i into other states
        norm_trans = np.zeros((K))

        for i in range(1, len(sorted_acts)):
            zn = sorted_acts[i-1]
            znp1 = sorted_acts[i]
            zn_i = self._hmm._idx_state(str(zn))
            zn_j = self._hmm._idx_state(str(znp1))
            tm[zn_i][zn_j] += 1
            norm_trans[zn_i] += 1
        #print(tm)
        #print(norm_trans)
        """
        count the amount of constants used in transition matrix
        this has to be evenly subtracted from the normalized probabilites
        
        """
        #zero_count = K**2 - np.count_nonzero(tm)
        #print(zero_count)
        #constant_correction = zero_count*self._transition_constant/ (K**2 - zero_count)
        #print(constant_correction)

        for i in range(0, K):
            row_zeros_count = K - np.count_nonzero(tm[i])
            if K == row_zeros_count:
                # if the complete has no activity set it to standard
                tm[i] = np.random.random_sample(K)
                tm[i] = tm[i] / sum(tm[i])
            else:
                row_correction =  row_zeros_count*self._transition_constant/ (K-row_zeros_count)
                for j in range(0, K):
                    if tm[i][j] == 0:
                        # set to nonzero constant
                        tm[i][j] = self._transition_constant
                    else:
                        # normalize with amount of outgoing transitions
                        tm[i][j] = (tm[i][j]/norm_trans[i]) - row_correction
        return tm



    def gen_pc_emissions(self, K, D):
        """
        :param K:
            the amount of states
        :param D:
            the amount of observations
        :return:
            K x D numpy array
        """
        act_dev_lst = self._gen_pc_em_transform_loc_data()
        #print('loc_data: ', str(self._loc_data))
        #print('dev_lst: ', act_dev_lst)
        #print('-'*10)

        em = np.zeros((K, D))
        #print('em: ', em)
        obs_list = self._hmm._o
        for k, act_dev in enumerate(act_dev_lst):
            #print('#'*100)
            #print('k: ', k)
            #print('act_dev_lst: ', act_dev)
            cnt_ones = len(act_dev)
            #cnt_ones = 0
            #for obs in obs_list:
            #    if obs in act_dev:
            #        cnt_ones += 1
            #print(cnt_ones)
            m = cnt_ones
            if cnt_ones == 0:
                # if there is no observation related to this activity
                # then set the observations for this activity to random values
                em[k] = np.random.random_sample(D)
                em[k] = em[k] / sum(em[k])
            else:
                prob_ones = (1 - (len(obs_list) - m) * self._emission_constant) / m
                #print('prob_ones: ', prob_ones)
                #print('c: ', self._c)
                #print()
                for d, obs in enumerate(obs_list):
                    if obs in act_dev:
                        em[k][d] = prob_ones
                    else:
                        em[k][d] = self._emission_constant
        return em

    def _gen_pc_em_transform_loc_data(self):
        act_state_list = self._hmm._z
        act_dev_lst = []
        #print('state_lst: ', act_state_list)
        #print('dev_lst: ', act_dev_lst)
        #print('loc_data: ', str(self._loc_data))
        """
        [[1,2,3], [2], ... ] 
        with 1,2,3 being the observations for activity 1 
        """
        for idx_z, z in enumerate(act_state_list):
            #print('#'*100)
            #print('act: ', z)
            #print('dev_lst: ', act_dev_lst)
            for loc in self._loc_data:
                #print('*'*10)
                #print('act: ', z)
                #print('lact: ', loc['activities'])
                if z in loc['activities'] and loc['devices'] != []:
                    #print('True')
                    try:
                        act_dev_lst[idx_z].extend(loc['devices'])
                    except:
                        # the first time the there is no list and therefore has
                        # to be assigned
                        #print('idx_z: ', str(idx_z))
                        act_dev_lst.append([])
                        act_dev_lst[idx_z].extend(loc['devices'])
            if len(act_dev_lst)-1 < int(z):
                """
                this is the case if no device was assigned to
                the activity, then an empty list has to be appended
                """
                act_dev_lst.append([])

        return act_dev_lst



from torch import nn
from pyadlml.model.nn import Tanh, Embedding, FlattenConsecutive
from torch.nn import Linear, BatchNorm1d
import torch 

def print_in_out(layer, name=None):
    def wrapper(x):
        i = list(x.shape)
        x = layer(x)
        j = list(x.shape)
        name = layer.__class__.__name__ if name is None else name
        print(f'{name} {i} -> {j}')
        return x
    return wrapper

class WaveNet(nn.Module):

    def __init__(self, n_features, n_emb, n_hidden, n_classes, dilation_factor):
        """
        """
        super().__init__()

        d = dilation_factor
        self.l_emb = Embedding(n_features, n_emb)             # B, T            -> B, T, E
        self.l_fltn1 = FlattenConsecutive(d)                   # B, T, E         -> B, T/d, E*d
        self.l_lin1 = Linear(n_emb*d, n_hidden, bias=False)    # B, T/d, E*d     -> B, T/d, H
        self.l_bn1 = BatchNorm1d(n_hidden)                     # B, T/d, H       -> B, T/d, H
        self.l_tanh1 = Tanh()                                  # B, T/d, H       -> B, T/d, H

        self.l_fltn2 = FlattenConsecutive(d)                      # B, T/d, H       -> B, T/d^2, H*d
        self.l_lin2 = Linear(n_hidden*d, n_hidden, bias=False) # B, T/d^2, H*d   -> B, T/d^2, H*d
        self.l_bn2 = BatchNorm1d(n_hidden)                     # B, T/d^2, H*d   -> B, T/d^2, H*d
        self.l_tnh2 = Tanh()                                   # B, T/d^2, H*d   -> B, T/d^2, H*d

        self.l_fltn3 = FlattenConsecutive(d)                      # B, T/d^2, H     -> B, T/d^3, H*d^2
        self.l_lin3 = Linear(n_hidden*d, n_hidden, bias=False) # B, T/d^3, H*d^2 -> B, T/d^3, H*d^2
        self.l_bn3 = BatchNorm1d(n_hidden)                     # B, T/d^3, H*d^2 -> B, T/d^3, H*d^2
        self.l_tnh3 = Tanh()                                   # B, T/d^3, H*d^2 -> B, T/d^3, H*d^2

        self.l_fltn4 = FlattenConsecutive(d)                      # B, T/d^2, H     -> B, T/d^3, H*d^2
        self.l_lin4 = Linear(n_hidden*d, n_hidden, bias=False) # B, T/d^3, H*d^2 -> B, T/d^3, H*d^2
        self.l_bn4 = BatchNorm1d(n_hidden)                     # B, T/d^3, H*d^2 -> B, T/d^3, H*d^2
        self.l_tnh4 = Tanh()                                   # B, T/d^3, H*d^2 -> B, T/d^3, H*d^2

        self.lin_5 = nn.Linear(n_hidden, n_classes)


    @property
    def layers(self):
        return self.parameters()


    def forward(self, x):
        """
        """



        x = self.l_emb(x)
        x = self.l_fltn1(x) 
        x = self.l_lin1(x)  # d -> B, T/d, H
        x = x.swapdims(1,2)
        # Expects (N, C, L) where N=batchsize, C=features, L = seq_length
        x = self.l_bn1(x)
        x = x.swapdims(1,2)
        x = self.l_tanh1(x)

        x = self.l_fltn2(x)
        x = self.l_lin2(x)
        x = x.swapdims(1,2)
        x = self.l_bn2(x)
        x = x.swapdims(1,2)
        x = self.l_tnh2(x)

        x = self.l_fltn3(x)
        x = self.l_lin3(x)
        x = x.swapdims(1,2)
        x = self.l_bn3(x)
        x = x.swapdims(1,2)
        x = self.l_tnh3(x)


        x = self.l_fltn4(x)
        x = self.l_lin4(x)
        #x = x.swapdims(1,2)
        x = self.l_bn4(x)
        #x = x.swapdims(1,2)
        x = self.l_tnh4(x)



        x = self.lin_5(x)
        assert x.dim() == 2, 'Bigrams do not resolve. Choose different numbers'

        return x





import plotly.graph_objects as go
import numpy as np
from plotly.subplots import make_subplots
import pandas as pd

from pyadlml.constants import ACTIVITY, STRFTIME_DATE
from pyadlml.dataset.plot.plotly.acts_and_devs import legend_current_items
from pyadlml.dataset.plot.plotly.util import remove_whitespace_around_fig, CatColMap

def rlencode(x, dropna=False):
    """
    Run length encoding.
    Based on http://stackoverflow.com/a/32681075, which is based on the rle 
    function from R.
    
    Parameters
    ----------
    x : 1D array_like
        Input array to encode
    dropna: bool, optional
        Drop all runs of NaNs.
    
    Returns
    -------
    start positions, run lengths, run values
    
    """
    where = np.flatnonzero
    x = np.asarray(x)
    n = len(x)
    if n == 0:
        return (np.array([], dtype=int), 
                np.array([], dtype=int), 
                np.array([], dtype=x.dtype))

    starts = np.r_[0, where(~(x[1:] == x[:-1])) + 1]
    lengths = np.diff(np.r_[starts, n])
    values = x[starts]
    
    if dropna:
        mask = ~np.isnan(values)
        starts, lengths, values = starts[mask], lengths[mask], values[mask]
    
    return starts, lengths, values


def confidences(y_conf, activities, cat_col_map=None):
    """
    Parameters
    ----------
    y_conf: nd.array of shape (N,C)
    classes: list or nd.array of shape (C,)
        Each class labels position corresponds to the confidence index in y_conf:
        c_i \in classes <=> (N, c_i) in y_conf

    """
    if cat_col_map is None:
        cat_col_map = CatColMap()
    assert isinstance(cat_col_map, CatColMap)

    fig = go.Figure()
    x=np.arange(0, len(y_conf))

    for c, act in enumerate(activities):
        cat_col_map.update(act, fig)
        y_c = y_conf[:,c]
        fig.add_trace(go.Scatter(
            x=x, y=y_c,
            mode='lines',
            name=act,
            line=dict(width=0.5, color=cat_col_map[act]),
            stackgroup='one'
        ))
    fig.update_layout(
        showlegend=True,
        yaxis=dict(
            type='linear',
            range=[0, 1],
        )
    )
    return fig


def _plot_confidences_into(fig, row, col, y_conf, activities, cat_col_map, 
                           time=None, step_fct=True

):
    if cat_col_map is None:
        cat_col_map = CatColMap()
    assert isinstance(cat_col_map, CatColMap)

    if time is not None:
        x = time
    else:
        x=np.arange(0, len(y_conf))

    # Fix display issues of not displaing low prob lines
    eps = 1e-6
    y_conf = np.where(y_conf == 0, y_conf + eps, y_conf)

    for c, act_name in enumerate(activities):
        cat_col_map.update(act_name, fig)
        showlegend = act_name not in legend_current_items(fig)
        y_c = y_conf[:,c]

        # Set the probabilities to a step function if needed
        line_dct = dict(width=0.5, color=cat_col_map[act_name])
        if step_fct:
            line_dct['shape'] = 'hv'

        hover_template = ''#'%{y:.4f}'
        for ci, act in enumerate(activities): 
            hover_template += f'<b>{act}</b>' if act == act_name else act
            hover_template +=  ': %{customdata[' + str(ci) + ']:.4f}<br>'
        cm = y_conf

        fig.add_trace(go.Scatter(
                x=x, y=y_c,
                mode='lines',
                name=act_name,
                customdata=cm, 
                hovertemplate=hover_template,
                line=line_dct,
                stackgroup='one',
                legendgroup=act_name,
                showlegend=showlegend,
            ), row=row, col=col)

    # Make plot not scrollabel
    fig.update_yaxes(row=1, col=1, fixedrange=True, range=[0, 1])
    return fig

 


def _plot_activities_into(fig, y, y_label, cat_col_map, row=1, col=1, time=None, activities=None):
        trace_lst = []

        if activities is None:
            activities = np.unique(y)

        df = pd.DataFrame(data=zip(*rlencode(y)), columns=['start', 'lengths', ACTIVITY])
        df['y_label'] = y_label
        if time is not None:
            df = discreteRle2timeRle(df, time)

        df['end'] = df['start'] + df['lengths']
        if time is not None:
            df['lengths' ] = df['lengths'].astype("timedelta64[ms]")

        set_legendgroup_title = 'Activities'
        for act_name in activities:
            cat_col_map.update(act_name, fig)
            show_legend = act_name not in legend_current_items(fig)
            act_mask = (df[ACTIVITY] == act_name)
            if time is None:
                hover_template = '<b>' + act_name + '</b><br>'\
                                + 'Int: [%{base}, %{x})<br>' \
                                + 'Dur: %{customdata}<extra></extra>'
            else:
                hover_template = '<b>' + act_name + '</b><br>'\
                                + 'Start_time: %{base|' + STRFTIME_DATE + '}<br>' \
                                + 'End_time: %{x| ' + STRFTIME_DATE + '}<br>' \
                                + 'Dur: %{customdata}<extra></extra>'

            trace = go.Bar(name=act_name,
                            meta=y_label,
                            base=df.loc[act_mask, 'start'],
                            x=df.loc[act_mask, 'lengths'],
                            y=df.loc[act_mask, 'y_label'],
                            legendgrouptitle_text=set_legendgroup_title,
                            marker_color=cat_col_map[act_name],
                            legendgroup=act_name,
                            customdata=df.loc[act_mask, 'end'],
                            orientation='h',
                            width=0.9,
                            textposition='auto',
                            alignmentgroup=True,
                            offsetgroup=act_name,
                            showlegend=show_legend,
                            hovertemplate=hover_template,
            )
            trace_lst.append(trace)
            set_legendgroup_title = None

        [fig.add_trace(trace, row=row, col=col) for trace in trace_lst]

        # Add additional 
        acts_not_in_legend = set(activities) - set(legend_current_items(fig))
        for act_name in acts_not_in_legend:
            # TODO check if it works
            raise NotImplementedError
            cat_col_map.update(act_name, fig)
            fig.add_trace(go.Bar(
                name=act_name, 
                visible='legendonly',
                showlegend=True
            ))

        # Add traces, layout and frames to figure
        fig.update_layout({'barmode': 'overlay',
                        'legend': {'tracegroupgap': 0}
                        })
        return fig

def discreteRle2timeRle(df, time):
    delta_t = (time - time.shift(1))[1:].reset_index(drop=True)
    df['start_time'] = time[0]
    df['lengths_time'] = pd.Timedelta('0ns')
    def func(row, t, dt):
        row.start_time = t[row.start]
        if row.lengths != 1:
            row.lengths_time = dt[row.start:row.start+row.lengths].sum()
        elif row.start != len(dt):
            row.lengths_time = dt[row.start]
        else:
            # row is last row with lengths == 1
            row.lengths_time = pd.Timedelta('1s')
        return row

    return df.apply(func, axis=1, args=(time, delta_t))\
           .drop(columns=['start', 'lengths'])\
           .rename(columns={'start_time':'start', 'lengths_time':'lengths'})



def _plot_devices_into(fig, X, cat_col_map, row, col, time):
    ON = 1
    OFF = 0
    # TODO refactor, generalize also for numerical labels
    # Run length encoding of 0s and 1s
    bars = {f:{OFF:[], ON:[]} for f in X.columns}
    for f in bars.keys():
        bars[f] = pd.DataFrame(data=zip(*rlencode(X[f])), columns=['start', 'lengths', 'state'])
        if time is not None:
            bars[f] = discreteRle2timeRle(bars[f], time)
            bars[f]['lengths'] = bars[f]['lengths'].astype("timedelta64[ms]")

    trace_lst = []
    for dev in X.columns:
        for k in [0, 1]:
            if time is None:
                hover_template = '<b>' + dev + '</b><br>'\
                                + 'Int: [%{base}, %{x})<br>' \
                                + 'Dur: %{customdata}<extra></extra>'
            else:
                hover_template = '<b>' + dev + '</b><br>'\
                                + 'Start_time: %{base|' + STRFTIME_DATE + '}<br>' \
                                + 'End_time: %{x| ' + STRFTIME_DATE + '}<br>' \
                                + 'Dur: %{customdata}<extra></extra>'

            vals = bars[dev].loc[(bars[dev]['state'] == k)]
            trace = go.Bar(name=str(k),
                            meta=dev,
                            base=vals['start'],
                            x=vals['lengths'],
                            y=np.full(vals.shape[0], dev),
                            #legendgrouptitle_text=set_legendgroup_title,
                            marker_color=cat_col_map[k],
                            #legendgroup=cat,
                            #customdata=cd,
                            orientation='h',
                            width=0.3,
                            textposition='auto',
                            showlegend=False,
                            hovertemplate=hover_template,
            )
            trace_lst.append(trace)

    [fig.add_trace(trace, row=row, col=col) for trace in trace_lst]

    fig.update_layout({'barmode': 'overlay',
                       'legend': {'tracegroupgap': 0}
    })

    if time is not None:
        fig.update_xaxes(type="date")

    return fig

@remove_whitespace_around_fig
def acts_and_devs(X, y_true=None, y_pred=None, y_conf=None, act_order=None, time=None):
    """ Plot activities and devices for already 

    """
    if isinstance(y_true, pd.Series)\
        or isinstance(y_true, pd.DataFrame):
        y_true = y_true.to_numpy().squeeze()
    
    assert (y_true is None or isinstance(y_true, np.ndarray))\
       and (y_pred is None or isinstance(y_pred, np.ndarray))\
       and (y_conf is None or isinstance(y_conf, np.ndarray))

    time = pd.Series(time) if isinstance(time, np.ndarray) else time

    error_text = 'parameter act_order has to be set if y_conf is given'
    assert y_conf is None or act_order is not None, error_text
    


    cat_col_map = CatColMap()
    X = X.copy().reset_index(drop=True)

    if y_conf is None:
        cols, rows, row_heights = 1, 1, [1.0]
    else:
        cols, rows, row_heights = 1, 2, [0.15, 0.85]

    fig = make_subplots(cols=cols, rows=rows, row_heights=row_heights,                        
        shared_xaxes=True, vertical_spacing=0.02,
    )

    fig = _plot_devices_into(fig, X, cat_col_map, row=rows, col=cols, time=time)

    if y_true is not None:
        fig = _plot_activities_into(fig, y_true, 'y_true', cat_col_map, row=rows, col=cols, time=time, activities=act_order)
    if y_pred is not None:
        fig = _plot_activities_into(fig, y_pred, 'y_pred', cat_col_map, row=rows, col=cols, time=time)
    if y_conf is not None:
        fig = _plot_confidences_into(fig, 1, 1, y_conf, act_order,
                                     cat_col_map,  time)

    return fig




import pandas as pd
from pyadlml.constants import TIME, DEVICE, VALUE
from pyadlml.dataset._representations.changepoint import create_changepoint
from pyadlml.dataset._core.acts_and_devs import label_data


def create_lastfired(df_devs):
    """
    creates the last fired representation
    """
    return create_changepoint(df_devs)


def resample_last_fired(lf, t_res):
    """

    Parameters
    ----------
    lf : pd.DataFrame
        last fired representation

    """
    lf = lf.set_index(TIME)
    resampler = lf.resample(t_res, kind='timestamp')
    lf = resampler.apply(_lf_evaluator, df=lf.copy())
    lf = lf.fillna(method='ffill')\
        .reset_index()
    return lf


def _lf_evaluator(series: pd.Series, df:pd.DataFrame) -> int:
    """
    Is called column-wise for each element of the
    
    Parameters
    ----------
    series : pd.Series
        conatins name of the column the evaluator operates on
        contains the timestamps for that change in that frame and the value
        that the specified column has at that timestamp.

        is empty if all columns that have no changing entity

    df : pd.DataFrame
        contains device representation 3. Is for determining the states of other
        device/columns for the timestamps

    Returns: 0|1
    """
    #print('series: ', series, '\n')
    if series.empty:
        return None
    else:         
        # check if this device is triggered during the timeslice
        # this is okay because other devices where triggered -> it has to be zero
        if series.sum() == 0:
            return 0
        
        # get dataframe for timeslice
        ts = series.index[0]
        if len(series) == 1:
            te = ts
        else:
            te = series.index[-1]
        df_slice = df.loc[ts:te]
        
        # if the device changed the state check if it was the 
        # last device to do so in the interval. When true 
        # the timeslice should be 1 for device last fired and 0 otherwise
        for i in range(len(df_slice)-1, -1, -1):
            
            # check if any device triggered
            if df_slice.iloc[i].values.sum() == 0:
                continue
                
            # check if our device was the one who triggered
            elif df_slice.loc[df_slice.index[i], series.name] == 1:
                return 1
            else:
                return 0



import numpy as np
import pandas as pd
from sklearn.model_selection._split import _BaseKFold

from pyadlml.constants import ACTIVITY, TIME, START_TIME, END_TIME, DEVICE, VALUE
from sklearn.model_selection import TimeSeriesSplit as SklearnTSSplit, KFold as SklearnKFold

from pyadlml.dataset.util import get_last_states, select_timespan
from pyadlml.dataset.cleaning.misc import remove_days


def train_test_split(df_devs, df_acts, split='leave_one_day_out', temporal=False,
                     return_day=False, return_init_states=False):
    """
    Splits the data into training and test set.

    Parameters
    ----------
    df_devs : pd.DataFrame
        todo
    df_acts : pd.DataFrame
        todo
    split : one of {'leave_one_day_out', (float, float), (float, float, float)}, default='leave_one_day_out'
        Determines the way the data is split into train and test set.
        leave_one_day_out
            one day is selected at random and used as test set. The emerging gap is closed by moving all
            succeeding events by one day into the past.
        tuple 
            First float is the train set ratio, second the ratio for the test set. All floats have to sum to one.
        triplet
            First is the ratio for the train set, second the ratio for the validation set, the 
            third is the ratio for the test. All floats have to sum to one.
    temporal : bool, default=False
        If true the split using the ratio is done with respect to time rather then to datapoints.
    return_day : bool, default=False
        If set, returns the day that is used for the test-set as fifth argument. Takes only
        an effect if `split=leave_one_day_out`.
    return_test_init_states : bool, default=False
        If set returns the last device values before the split. These values are useful when generating
        StateVectors for the test set.

    Examples
    --------
    .. code python:
        from pyadlml.dataset import fetch_amsterdam
        from pyadlml.model_selection import train_test_split

        data = fetch_amsterdam()
        df_devs, df_acts = data['devices'], data['activities']

        X_train, X_test, y_train, y_test = train_test_split(
            df_devs, df_acts, split='leave_one_day_out'
        )
        

    Returns
    -------
    X_train, X_test, y_train, y_test : all pd.DataFrames

    """
    SPLIT_LODA = 'leave_one_day_out'
    TD = '1D'
    epsilon = pd.Timedelta('1ns')
    is_train_val_test_split = isinstance(split, tuple) and len(split) == 3
    is_train_test_split = not is_train_val_test_split
    ratio_split = isinstance(split, tuple)

    if ratio_split and isinstance(split, tuple):
        for elem in split:
            assert isinstance(elem, float) and 0 < elem and elem < 1.0
        assert sum(split) == 1
        assert len(split) == 2 or len(split) == 3
    else:
        assert split == SPLIT_LODA

    assert isinstance(temporal, bool) & isinstance(return_day, bool) & isinstance(return_init_states, bool)
    assert not (ratio_split and return_day)

    # just to make sure
    df_devs = df_devs.sort_values(by=TIME).reset_index(drop=True)

    if split == SPLIT_LODA:
        # Split by leaving one day out
        rnd_day = _get_rnd_day(df_devs, padding=1)
        X_train, y_train = remove_days(df_devs, df_acts, days=[rnd_day])
        X_test, y_test = select_timespan(df_devs, df_acts, 
                start_time=rnd_day, end_time=rnd_day + pd.Timedelta(TD),
                clip_activities=True
        )
    elif len(split) == 3:
        # Split by fraction 
        if temporal:
            total_range = (df_devs.at[len(df_devs)-1, TIME] - df_devs.at[0, TIME]) 
            train_range = total_range * split[0]
            val_range = total_range * (split[0] + split[1])

            train_end_time = df_devs.at[0, TIME] + train_range
            val_end_time = df_devs.at[0, TIME] + val_range
        else:
            train_end_time = df_devs.at[int(len(df_devs) * split[0]), TIME]
            val_end_time = df_devs.at[int(len(df_devs) * (split[0] + split[1])), TIME]

        X_train, y_train = select_timespan(df_devs, df_acts, 
               end_time=train_end_time - epsilon, clip_activities=True
        )
        X_val, y_val = select_timespan(df_devs, df_acts, 
               start_time=train_end_time + epsilon, 
               end_time=val_end_time - epsilon, clip_activities=True
        )
        X_test, y_test = select_timespan(df_devs, df_acts, 
                start_time=val_end_time + epsilon, 
                clip_activities=True
        )


    else:
        # Split by fraction 
        if temporal:
            train_range = (df_devs.at[len(df_devs)-1, TIME] - df_devs.at[0, TIME]) * split[0]
            split_time = df_devs.at[0, TIME] + train_range
        else:

            split_idx = int(len(df_devs) * split[0])
            split_time = df_devs.at[split_idx, TIME]

        X_train, y_train = select_timespan(df_devs, df_acts, 
               end_time=split_time - epsilon, clip_activities=True
        )
        X_test, y_test = select_timespan(df_devs, df_acts, 
                start_time=split_time + epsilon, clip_activities=True
        )


    if is_train_val_test_split:
        res_lst = [X_train, X_val, X_test, y_train, y_val, y_test]
    else:
        res_lst = [X_train, X_test, y_train, y_test]

    if return_day:
        res_lst.append([rnd_day, rnd_day + pd.Timedelta(TD)])

    if return_init_states:
        last_states_train = get_last_states(X_train)    

        if is_train_val_test_split:
            last_states_val = get_last_states(X_val)
            for dev in last_states_train.keys():
                if dev not in last_states_val.keys():
                    last_states_val[dev] = last_states_train[dev]
        
        if is_train_val_test_split:
            states_dict = dict(
                init_states_val=last_states_train,
                init_states_test=last_states_val
            )
        else:
            states_dict = dict(
                init_states_test=last_states_train,
            )
        
        res_lst.append(states_dict)


    # Sanity checks
    # 1. check if activity or device appears in test or val set that is not present in the train set
    if is_train_val_test_split:
        dev_in_train_but_not_in_val =  set(X_train[DEVICE].unique()) - set(X_val[DEVICE].unique())
        dev_in_val_but_not_in_train =  set(X_val[DEVICE].unique()) - set(X_train[DEVICE].unique())
        if dev_in_train_but_not_in_val:
            print('Warning: Devices from train set %s produce no event in validation set!'%dev_in_train_but_not_in_val)
        assert not dev_in_val_but_not_in_train, 'Unseen device in validation set not contained in train set %s. Data cleaning required!!!'%dev_in_val_but_not_in_train

    dev_in_train_but_not_in_test =  set(X_train[DEVICE].unique()) - set(X_test[DEVICE].unique())
    dev_in_test_but_not_in_train =  set(X_test[DEVICE].unique()) - set(X_train[DEVICE].unique())
    if dev_in_train_but_not_in_test:
        print('Warning: Devices from train set %s produce no event in test set!'%dev_in_train_but_not_in_test)
    assert not dev_in_test_but_not_in_train, 'Unseen device in test set not contained in train set %s. Data cleaning required!!!'%dev_in_test_but_not_in_train


    if is_train_val_test_split:
        act_in_train_but_not_in_val =  set(y_train[ACTIVITY].unique()) - set(y_val[ACTIVITY].unique())
        act_in_val_but_not_in_train =  set(y_val[ACTIVITY].unique()) - set(y_val[ACTIVITY].unique())
        if act_in_train_but_not_in_val:
            print('Warning: Activities from train set %s do not appear in validation set!'%act_in_train_but_not_in_val)
        assert not act_in_val_but_not_in_train, 'Unseen activity in validation set, not contained in train set %s. Data cleaning required!!!'%act_in_val_but_not_in_train


    act_in_train_but_not_in_test =  set(y_train[ACTIVITY].unique()) - set(y_test[ACTIVITY].unique())
    act_in_test_but_not_in_train =  set(y_test[ACTIVITY].unique()) - set(y_train[ACTIVITY].unique())
    if act_in_train_but_not_in_test:
        print('Warning: Activities from train set %s do not appear in test set!'%act_in_train_but_not_in_test)
    assert not act_in_test_but_not_in_train, 'Unseen activity in test set, not contained in train set %s. Data cleaning required!!!'%act_in_test_but_not_in_train



    return res_lst


def _get_rnd_day(df_devs, retain_other_days=False, padding=0):
    """ Retrieves a random day from the dataset

    Parameters
    ----------
    X : pd.DataFrame
        with timeindex
    retain_other_days : bool, default=False
        determines whether all other days except for the random day are also returned
    padding : int, default=0
        How many days from start and from end should not be used

    Returns
    -------
    str or list

    """

    # get all days
    days = list(df_devs[TIME].dt.floor('d').value_counts().index)
    days = days[padding:len(days)-padding-1]

    # select uniformly a random day
    rnd_idx = np.random.randint(0, high=len(days)-1)
    rnd_day = days[rnd_idx]
    if retain_other_days:
        return rnd_day, days.pop(rnd_idx)
    else:
        return rnd_day


def _split_devs(df_devs, rnd_day, temporal=False):
    """ get indices of all data for that day and the others """
    if temporal:
        raise NotImplementedError
    else:
        rnd_dayp1 = rnd_day + pd.Timedelta('1D')
        mask = (rnd_day < df_devs[TIME]) & (df_devs[TIME] < rnd_dayp1)
        idxs_test = df_devs[mask].index.values
        idxs_train = df_devs[~mask].index.values
        return idxs_train, idxs_test


def _split_acts(df_acts : pd.DataFrame, rnd_day, temporal : bool =False) -> list:
    """ Get indices of all activities for that day and the others 
    
    Parameters
    ----------
    df_acts : pd.DataFrame
    temporal : 

    """
    if temporal:
        raise NotImplementedError
    else:
        rnd_dayp1 = rnd_day + pd.Timedelta('1D')
        mask_test = (rnd_day < df_acts[END_TIME]) & (df_acts[START_TIME] < rnd_dayp1)
        mask_train = (df_acts[START_TIME] < rnd_day) | (rnd_dayp1 < df_acts[END_TIME])
        idxs_test = df_acts[mask_test].index.values
        idxs_train = df_acts[mask_train].index.values
        return idxs_train, idxs_test


class TimeSeriesSplit(_BaseKFold):
    """
    Parameters
    ----------
    n_splits : int, default=5
        number of splits. Must be at least 2.
    max_train_size : int, default=None
        Maximum size for a single training set.
    test_size : int, default=None
        Used to limit the size of the test set. Defaults to n_samples // (n_splits + 1), which is the maximum allowed value with gap=0.
    gap : int, default=0
        Number of samples to exclude from the end of each train set before the test set.
    return_timestamp : bool, default=False
        When true timestamp intervals are returned rather than indicies. This is
        useful whenever data is upscaled or downscaled as the indicies in the testset c
        can not be known beforehand.
    temporal_split : bool, default=False
        If set, the splits are made based on the time rather than on the datapoints. This
        allows for rescaling of the data and applying the split afterwards.

    Examples
    --------
    >>> import os

    """
    EPS = pd.Timedelta('5ms')

    def __init__(self, n_splits=5, *, max_train_size=None, test_size=None, gap=0, return_timestamp=False,
                 temporal_split=False):
        self.return_timestamp = return_timestamp
        self.temporal_split = temporal_split
        self.max_train_size = max_train_size
        self.test_size = test_size

        if self.temporal_split:
            self.gap = pd.Timedelta('0s') if gap == 0 else gap
            assert isinstance(self.gap, pd.Timedelta)
        else:
            self.gap = gap

        super().__init__(n_splits, shuffle=False, random_state=None)


    def _temporal_split(self, X, y, groups):
        # create time_range from first device to last device
        assert isinstance(self.gap, pd.Timedelta)
        assert self.max_train_size is None or isinstance(self.max_train_size, pd.Timedelta)
        assert self.test_size is None or isinstance(self.test_size, pd.Timedelta)

        data_start = X[TIME].iloc[0]
        data_end = X[TIME].iloc[-1]
        n_folds = self.n_splits + 1 # |--|--|--|--|  k=3
        test_size = self.test_size if self.test_size is not None \
            else (data_end - data_start) // n_folds

        test_starts = pd.date_range(data_end - self.n_splits*test_size, data_end, freq=test_size)[:-1]

        res_lst = []
        for test_st in test_starts:
            train_et = test_st - self.gap - self.EPS
            test_et = test_st + test_size

            if self.max_train_size and self.max_train_size < train_et:
                train_st = train_et - self.max_train_size
            else:
                train_st = data_start - self.EPS

            if self.return_timestamp:
                res_lst.append(((train_st, train_et), (test_st, test_et)))
            else:
                train_idx = X[(train_st < X[TIME]) & (X[TIME] < train_et)].index.values
                test_idx = X[(test_st < X[TIME]) & (X[TIME] < test_et)].index.values
                res_lst.append((train_idx, test_idx))
        return res_lst


    def _index_split(self, X, y, groups):
        """ Blatantly copied from the original sklearn Timeseries split
        """
        assert isinstance(self.gap, int)
        assert self.max_train_size is None or isinstance(self.max_train_size, int)
        assert self.test_size is None or isinstance(self.test_size, int)

        n_samples = len(X)
        n_splits = self.n_splits
        n_folds = n_splits + 1
        gap = self.gap
        test_size = self.test_size if self.test_size is not None \
            else n_samples // n_folds

        # Make sure we have enough samples for the given split parameters
        if n_folds > n_samples:
            raise ValueError(
                (f"Cannot have number of folds={n_folds} greater"
                 f" than the number of samples={n_samples}."))
        if n_samples - gap - (test_size * n_splits) <= 0:
            raise ValueError(
                (f"Too many splits={n_splits} for number of samples"
                 f"={n_samples} with test_size={test_size} and gap={gap}."))

        indices = np.arange(n_samples)
        test_starts = range(n_samples - n_splits * test_size,
                            n_samples, test_size)
        res_lst = []
        for test_start in test_starts:
            train_end = test_start - gap
            if self.max_train_size and self.max_train_size < train_end:
                train_idxs = indices[train_end - self.max_train_size:train_end]  # sliding window
            else:
                train_idxs = indices[:train_end]  # expanding window
            test_idxs = indices[test_start:test_start + test_size]

            # own implementation addition
            if not self.return_timestamp:
                res_lst.append((train_idxs, test_idxs))
            else:
                train_st = X.iloc[train_idxs[0]][TIME] - self.EPS
                train_et = X.iloc[train_idxs[-1]][TIME] + self.EPS

                val_st = X.iloc[test_idxs[0]][TIME] - self.EPS
                val_et = X.iloc[test_idxs[-1]][TIME] + self.EPS

                res_lst.append(
                    ((train_st, train_et), (val_st, val_et))
                )

        return res_lst


    def split(self, X, y=None, groups=None):
        """ Generate indices or intervals to split data into training and test set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        y : array-like of shape (n_samples,)
            Always ignored, exists for compatibility.

        groups : array-like of shape (n_samples,)
            Always ignored, exists for compatibility.

        Returns
        -------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.

        """
        if self.temporal_split:
            return self._temporal_split(X, y, groups)
        else:
            return self._index_split(X, y, groups)


class LeaveKDayOutSplit(object):
    """ LeaveKDayOut cross-validator

    Provides train/test indices to split data in train/test sets. Split
    dataset into one day out folds.

    Read more in the :ref:`User Guide <leave_one_day_out>`

    Parameters
    ----------
    k : int, default=1
        The number of days to use for the test set.
    n_splits : int, default=1
        The number of splits. All splits are exclusive, meaning there will not be more t TODO
    return_timestamps : bool, default=False
        When true timestamp intervals are returned rather than indicies. This is
        useful whenever data is upscaled or downscaled as the indicies in the testset
        can not be known beforehand.
    epsilon : str, default='5ms'
        the offset that is used to pad before the first and after the last interval for
        the timestamps. Has only an effect if *return_timestamps* is set to *true*
    offset : str, default='0s'
        The offset that is used to shift the start of a day
    shift : bool, defaul=False
        Determines whether to shift the

    Examples
    --------
    >>> import os
    """
    def __init__(self, n_splits=1, k=1, return_timestamps=False, epsilon='5ms', offset='0s', shift=False):
        self.n_splits = n_splits
        self.return_timestamp = return_timestamps
        self.k = k
        self.offset = pd.Timedelta(offset)
        self.eps = pd.Timedelta(epsilon)
        self.shift = shift

    def get_n_splits(self, X=None, y=None, groups=None):
        """Returns the number of splitting iterations in the cross-validator
        Parameters
        ----------
        X : object
            Always ignored, exists for compatibility.
        y : object
            Always ignored, exists for compatibility.
        groups : array-like of shape (n_samples,)
            Group labels for the samples used while splitting the dataset into
            train/test set. This 'groups' parameter must always be specified to
            calculate the number of splits, though the other parameters can be
            omitted.
        Returns
        -------
        n_splits : int
            Returns the number of splitting iterations in the cross-validator.
        """
        return self.n_splits

    def split(self, X=None, y=None, groups=None):
        """ Generate indices to split data into training and test set.

        Parameters
        ----------
        X : pd.DataFrame
            device dataframe
        y : pd.Series
            activity dataframe

        Returns
        -------
        splits : list
            Returns tuples of splits of train and test sets
            example: [(train1, test1), ..., (trainn, testn)]
        """

        X = X.copy()
        first_day = X[TIME].iloc[0].floor('d')
        last_day = X[TIME].iloc[-1].ceil('d')
        days = pd.date_range(first_day, last_day, freq='1D').values
        print(days)
        days[1:-2] = days[1:-2] + self.offset

        N = len(days)
        if self.k is None:
            self.k = (N-2)//self.n_splits

        assert self.k <= (N-2)//self.n_splits, "The number of days for each split exceeds the possible"

        step_size = N//self.n_splits
        res = []
        for i in range(self.n_splits):
            test = (days[i*step_size], days[i*step_size + self.k])
            if i == 0:
                # case when | test | train |
                train = (days[i*step_size + self.k], days[-1])
            elif i == self.n_splits-1 and (i*step_size+self.k) == N-1:
                # case  when | train | test|
                train = (days[0], days[i*step_size])
            else:
                # case when | train | test | train |
                train = ((days[0], days[i*step_size]),
                         (days[i*step_size + self.k], days[-1]))

            if self.return_timestamp:
                res.append((train, test))
            else:
                def get_indices(df, l_bound, r_bound):
                    return df[(l_bound < df[TIME]) & (df[TIME] < r_bound)].index.values
                test_idxs = get_indices(X, test[0], test[1])
                if i == 0 or (i == self.n_splits-1 and (i*step_size+self.k) == N-1):
                    train_idxs = get_indices(X, train[0], train[1])
                else:
                    train_idxs_int_1 = get_indices(X, train[0][0], train[0][1])
                    train_idxs_int_2 = get_indices(X, train[1][0], train[1][1])
                    train_idxs = np.concatenate([train_idxs_int_1, train_idxs_int_2])

                    if self.shift:
                        # shift the second interval by that amount of days into the past
                        X.loc[train_idxs_int_2, TIME] = X[TIME] - pd.Timedelta(str(self.k) + 'D')
                        if y is not None:
                            y.loc[train_idxs_int_2, TIME] = y[TIME] - pd.Timedelta(str(self.k) + 'D')

                res.append((train_idxs, test_idxs))

        return res


from pyadlml.dataset.plot.util import heatmap, annotate_heatmap
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.pyplot as plt
import matplotlib
import numpy as np

def forceAspect(ax,aspect):
    im = ax.get_images()
    extent =  im[0].get_extent()
    ax.set_aspect(abs((extent[1]-extent[0])/(extent[3]-extent[2]))/aspect)
    
def mean_image(images, devices, figsize=(10,10)):
    """ plots the mean oimage of all images
    Parameters
    ----------
    images: 3d np array (n x win_size x dev_count)
        a 3d batch of images with window size and the feature/devices as last axis
    devices: list 
        a list of all device labels
    
    """
    title='Mean image'
    
    mean_image = images.mean(axis=0).T
    
    fig, ax = plt.subplots(figsize=figsize)
    h, w = mean_image.shape
    
    im = ax.imshow(mean_image, extent=[0,w,0,h], vmin=0, vmax=1)
    
    
    divider = make_axes_locatable(ax)      
    cax = divider.append_axes("right", size='5%', pad=0.1)
    cbar = cax.figure.colorbar(im, cax)
    #plt.colorbar(im, cax=cax)
    
    
    ax.set_xticks(np.arange(w) + 0.5)
    ax.set_yticks(np.arange(h) + 0.5)
    
    ax.set_xticklabels(np.arange(w)+1)
    ax.set_yticklabels(devices[::-1]) # reverse because of transpose of values
    
    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
         rotation_mode="anchor")

    
    #forceAspect(ax, aspect=1.5)
    ax.set_title(title)
    fig.tight_layout()
    plt.show()



def mean_image_per_activity(X, y, devices, figsize=(14,18)):
    """ plots the mean oimage of all images
    Parameters
    ----------
    X: 3d np array (n x win_size x dev_count)
        a 3d batch of images with window size and the feature/devices as last axis
    y: 1d array (n,) of strings
        activity labels 
    devices: list 
        a list of all device labels
    """
    title='Mean image per Activity'
    num_plts_x = 3
    
    activities = np.unique(y)
    
    mi_lst = []
    for act in activities:
        # get the index where the activity matches and compute mean
        idxs = np.where(y == act)[0]    
        mi_lst.append(X[idxs].mean(axis=0).T)
    
    len_acts = len(activities)
    num_plts_y = int(np.ceil(len_acts/num_plts_x))
    
    
    # plotting
    fig, axs = plt.subplots(num_plts_y, num_plts_x, 
                             figsize=figsize)
    #plt.title('asdf', y=1.08)
    
    h, w = mi_lst[0].shape
    k = 0
    pcm_list = []
    for i in range(num_plts_y):
        for j in range(num_plts_x):
            if k >= len(activities):
                axs[i,j].axis('off')
                continue
                
            ax = axs[i,j]
            im = ax.imshow(mi_lst[k],  vmin=0, vmax=1)#, extent=[0,w,0,h],)
            #pcm = ax.pcolormesh(mi_lst[k], cmap='viridis')
            #pcm_list.append(pcm)
            
            #if k+1 == len(activities):
            ax.set_yticks(np.arange(h))
            if j == 0:                                
                ax.set_yticklabels(devices)
            else:
                ax.set_yticklabels([])
            ax.set_xticks(np.arange(w))
            ax.set_xticklabels(np.arange(w)+1)
            
            ax.set_title(activities[k])
            k += 1
    
    # make cosmetic changes
    plt.subplots_adjust(wspace=0.1, hspace=-.6)
    plt.suptitle(title,fontsize=20, y=0.8, va='top')
    plt.show()